import logging
from typing import Any, Callable

from ..storage import WebmentionsStorage
from .._model import (
    ContentTextFormat,
    Webmention,
    WebmentionDirection,
    WebmentionStatus,
)
from ._constants import DEFAULT_HTTP_TIMEOUT, DEFAULT_USER_AGENT
from ._incoming import IncomingWebmentionsProcessor
from ._outgoing import OutgoingWebmentionsProcessor

logger = logging.getLogger(__name__)


class WebmentionsHandler:
    """
    Webmentions handler.

    :param storage: The Webmentions storage backend
    :param base_url: The base URL of the server, used to validate target URLs
    :param http_timeout: The HTTP timeout for fetching source URLs
    :param user_agent: The User-Agent header to use when fetching source URLs
    :param on_mention_processed: A callback to call when a Webmention is processed
    :param on_mention_deleted: A callback to call when a Webmention is deleted
    :param initial_mention_status: The initial status of Webmentions (see
        :class:`WebmentionStatus`). If not specified, defaults to
        :attr:`WebmentionStatus.CONFIRMED`. If you set this to
        :attr:`WebmentionStatus.PENDING` then you will need to manually mark
        mentions as confirmed on your storage, or create your on
        ``on_mention_processed`` that performs custom filtering or moderation
        and calls ``handler.storage.store_webmention(webmention)`` with the right
        status after processing.
    """

    def __init__(
        self,
        storage: WebmentionsStorage,
        *,
        base_url: str | None = None,
        http_timeout: float = DEFAULT_HTTP_TIMEOUT,
        user_agent: str = DEFAULT_USER_AGENT,
        on_mention_processed: Callable[[Webmention], None] | None = None,
        on_mention_deleted: Callable[[Webmention], None] | None = None,
        initial_mention_status: WebmentionStatus = WebmentionStatus.CONFIRMED,
        **kwargs,
    ):
        self.storage = storage
        self.incoming = IncomingWebmentionsProcessor(
            storage=storage,
            base_url=base_url,
            http_timeout=http_timeout,
            user_agent=user_agent,
            on_mention_processed=on_mention_processed,
            on_mention_deleted=on_mention_deleted,
            init_mention_status=initial_mention_status,
            **kwargs,
        )
        self.outgoing = OutgoingWebmentionsProcessor(
            storage=storage,
            user_agent=user_agent,
            http_timeout=http_timeout,
            on_mention_processed=on_mention_processed,
            on_mention_deleted=on_mention_deleted,
            **kwargs,
        )

    def process_incoming_webmention(
        self, source_url: str | None, target_url: str | None
    ) -> Any:
        """
        Process an incoming Webmention.

        :param source_url: The source URL of the Webmention
        :param target_url: The target URL of the Webmention
        """
        return self.incoming.process_incoming_webmention(source_url, target_url)

    def process_outgoing_webmentions(
        self,
        source_url: str,
        *,
        text: str | None = None,
        text_format: ContentTextFormat | None = None,
    ) -> Any:
        """
        Process an outgoing Webmention.

        :param source_url: The source URL of the Webmention. Ignored if text is
            provided.
        :param text: The text of the Webmention. If not provided, the source URL
            will be fetched.
        :param text_format: The text format of the Webmention. If not provided,
            it will be inferred from the source URL or text.
        """
        return self.outgoing.process_outgoing_webmentions(
            source_url, text=text, text_format=text_format
        )

    def retrieve_stored_webmentions(
        self, resource: str, direction: WebmentionDirection
    ) -> list[Webmention]:
        """
        Retrieve stored Webmentions for a given URL.

        :param resource: The resource URL
        :param direction: The direction of the Webmentions (inbound or outbound)
        :return: A list of Webmentions
        """
        return self.storage.retrieve_webmentions(resource, direction=direction)



import sqlalchemy as sa
from sqlalchemy.orm import declarative_base
from sqlalchemy.orm import sessionmaker

from ._model import DbWebmention
from ._storage import DbWebmentionsStorage


def init_db_storage(
    engine: str | sa.Engine,
    *args,
    table_name: str = "webmentions",
    **kwargs,
) -> DbWebmentionsStorage:
    """
    A helper function that initializes a database storage for Webmentions.

    Use this if you want to use a dedicated SQLAlchemy engine for your
    Webmentions storage.

    Otherwise, extend ``DbWebmention`` with your table name, register it
    to your engine, and initialize a ``DbWebmentionsStorage``.

    :param engine: SQLAlchemy engine, as a string (creates a new engine)
        or an existing engine
    :param table_name: Name of the table to use to store Webmentions
    :param args: Positional arguments to pass to ``sa.create_engine``
    :param kwargs: Keyword arguments to pass to ``sa.create_engine``
    """

    Base = declarative_base()

    class DefaultDbWebmention(Base, DbWebmention):
        """
        Dynamically generated base class.
        """

        __tablename__ = table_name

    if isinstance(engine, str):
        engine = sa.create_engine(engine, *args, **kwargs)

    Base.metadata.create_all(engine)
    return DbWebmentionsStorage(
        engine=engine,
        session_factory=sessionmaker(bind=engine),
        model=DefaultDbWebmention,
    )



import flask as flask_upstream  # pylint: disable=W0406

if getattr(flask_upstream, "__file__", None) == __file__:
    raise RuntimeError(
        "Local module name 'flask.py' is shadowing the upstream 'flask' dependency. "
        "Do not run this file directly; import it as 'webmentions.server.adapters.flask'."
    )

Flask = flask_upstream.Flask
Blueprint = flask_upstream.Blueprint
jsonify = flask_upstream.jsonify
request = flask_upstream.request

from ...handlers import WebmentionsHandler
from ..._exceptions import WebmentionException
from ..._model import WebmentionDirection
from ._common import append_link_header, webmention_link_header_value


def _join_url_prefix(url_prefix: str | None, route: str) -> str:
    prefix = (url_prefix or "").rstrip("/")
    path = route if route.startswith("/") else f"/{route}"
    if not prefix:
        return path
    return f"{prefix}{path}"


def _install_webmentions_link_header_after_request(app: Flask):
    if not app.extensions.get("webmentions_link_header_after_request_installed", False):
        app.extensions["webmentions_link_header_after_request_installed"] = True

        @app.after_request
        def _webmentions_link_header_after_request(response):
            content_type = response.headers.get("Content-Type")
            if content_type is not None and content_type.split(";", 1)[
                0
            ].strip().startswith("text/"):
                existing = response.headers.get("Link")
                for endpoint in app.extensions.get("webmentions_endpoints", set()):
                    existing = append_link_header(
                        existing, webmention_link_header_value(endpoint)
                    )
                if existing is not None:
                    response.headers["Link"] = existing
            return response

        return _webmentions_link_header_after_request

    return None


def webmention_route(handler: WebmentionsHandler):
    source = request.form.get("source")
    target = request.form.get("target")

    try:
        handler.process_incoming_webmention(source, target)
    except WebmentionException as e:
        return jsonify({"error": str(e)}), 400

    return jsonify({"status": "ok"})


def retrieve_webmentions_route(handler: WebmentionsHandler):
    resource = request.args.get("resource")
    direction_raw = request.args.get("direction")
    if not resource:
        return jsonify({"error": "resource is required"}), 400
    if not direction_raw:
        return jsonify({"error": "direction is required"}), 400

    try:
        direction = WebmentionDirection.from_raw(direction_raw)
    except ValueError:
        return jsonify({"error": "invalid direction"}), 400

    try:
        stored = handler.storage.retrieve_webmentions(resource, direction=direction)
    except Exception as e:
        return jsonify({"error": str(e)}), 400

    return jsonify([wm.to_dict() for wm in stored])


def bind_webmentions(
    app: Flask, handler: "WebmentionsHandler", route: str = "/webmentions"
):
    """
    Bind a Flask endpoint to process incoming Webmentions.

    :param app: The Flask application to bind the endpoint to.
    :param handler: The WebmentionsHandler to use for processing incoming Webmentions.
    :param route: The route to bind the endpoint to.
    """

    app.extensions.setdefault("webmentions_endpoints", set()).add(route)
    _install_webmentions_link_header_after_request(app)

    def _route():
        return webmention_route(handler)

    def _get_route():
        return retrieve_webmentions_route(handler)

    endpoint_name = f"webmention_{abs(hash(route))}"
    app.add_url_rule(route, endpoint=endpoint_name, view_func=_route, methods=["POST"])

    endpoint_get_name = f"webmention_get_{abs(hash(route))}"
    app.add_url_rule(
        route, endpoint=endpoint_get_name, view_func=_get_route, methods=["GET"]
    )
    return route


def bind_webmentions_blueprint(
    handler: "WebmentionsHandler", route: str = "/webmentions"
) -> Blueprint:
    """
    Create a Flask Blueprint with a bound Webmentions endpoint.

    :param handler: The WebmentionsHandler to use for processing incoming Webmentions.
    :param route: The route to bind the endpoint to.
    :return: A Flask Blueprint instance.
    """

    bp = Blueprint("webmentions", __name__)

    @bp.record_once
    def _register(state):
        app = state.app
        effective_route = _join_url_prefix(state.url_prefix, route)
        app.extensions.setdefault("webmentions_endpoints", set()).add(effective_route)
        _install_webmentions_link_header_after_request(app)

    @bp.post(route)
    def _route():
        return webmention_route(handler)

    return bp



from dataclasses import asdict
from datetime import datetime, timezone
from typing import Callable

import sqlalchemy as sa
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import Session
from webmentions._model import WebmentionStatus

from ..._base import (
    Webmention,
    WebmentionDirection,
    WebmentionsStorage,
)
from ._model import DbWebmention


class DbWebmentionsStorage(WebmentionsStorage):
    """
    Implements a simple database storage for Webmentions.

    :param engine: SQLAlchemy engine
    :param model: SQLAlchemy model. It must inherit from DbWebmention
        and be mapped to the database.
    :param session_factory: SQLAlchemy session factory. You can create
        a session factory through e.g.

        .. code-block:: python

            from sqlalchemy.orm import sessionmaker

            session = sessionmaker(bind=engine)

    """

    def __init__(
        self,
        engine: sa.Engine,
        model: type[DbWebmention],
        session_factory: Callable[[], Session],
        *_,
        **__,
    ):
        self.engine = engine
        self.session_factory = session_factory
        self.model = model

    def store_webmention(self, mention: Webmention):
        session = self.session_factory()
        try:
            try:
                session.add(self.model.from_webmention(mention))
                session.commit()
                return
            except IntegrityError:
                session.rollback()

            existing = (
                session.query(self.model)
                .filter(
                    sa.and_(
                        self.model.source == mention.source,
                        self.model.target == mention.target,
                        self.model.direction == mention.direction,
                    )
                )
                .one_or_none()
            )

            if existing is None:
                session.add(self.model.from_webmention(mention))
                session.commit()
                return

            columns = self.model.columns()
            data = {k: v for k, v in asdict(mention).items() if k in columns}
            for key, value in data.items():
                if key in {"id", "created_at", "updated_at"}:
                    continue
                setattr(existing, key, value)

            existing.updated_at = datetime.now(timezone.utc)  # type: ignore
            session.commit()
        finally:
            session.close()

    def delete_webmention(
        self,
        source: str,
        target: str,
        direction: WebmentionDirection,
    ):
        session = self.session_factory()
        try:
            session.query(self.model).filter(
                sa.and_(
                    self.model.source == source,
                    self.model.target == target,
                    self.model.direction == direction,
                )
            ).delete(synchronize_session=False)
            session.commit()
        finally:
            session.close()

    def retrieve_webmentions(
        self,
        resource: str,
        direction: WebmentionDirection,
    ) -> list[Webmention]:
        session = self.session_factory()
        try:
            resource_column = (
                self.model.source
                if direction == WebmentionDirection.OUT
                else self.model.target
            )
            return [
                wm.to_webmention()
                for wm in session.query(self.model).filter(
                    sa.and_(
                        resource_column == resource,
                        self.model.direction == direction,
                        self.model.status == WebmentionStatus.CONFIRMED,
                    )
                )
            ]
        finally:
            session.close()



from .handlers import WebmentionsHandler
from .storage import WebmentionsStorage
from ._exceptions import WebmentionException, WebmentionGone
from ._model import (
    ContentTextFormat,
    Webmention,
    WebmentionDirection,
    WebmentionStatus,
    WebmentionType,
)

__version__ = "0.1.8"

__all__ = [
    "ContentTextFormat",
    "Webmention",
    "WebmentionDirection",
    "WebmentionException",
    "WebmentionGone",
    "WebmentionsHandler",
    "WebmentionsStorage",
    "WebmentionStatus",
    "WebmentionType",
]



from abc import ABC, abstractmethod
from datetime import datetime, timezone
from typing import Any

from .._model import Webmention, WebmentionDirection


class WebmentionsStorage(ABC):
    """
    Base class for Webmention storage backends.
    """

    @abstractmethod
    def store_webmention(self, mention: Webmention) -> Any:
        """
        Store a Webmention.

        :param mention: The Webmention to store
        """

    @abstractmethod
    def delete_webmention(
        self,
        source: str,
        target: str,
        direction: WebmentionDirection,
    ) -> Any | None:
        """
        Mark a Webmention as deleted.

        :param source: The source URL of the Webmention
        :param target: The target URL of the Webmention
        :param direction: The direction of the Webmention (inbound or outbound)
        """

    @abstractmethod
    def retrieve_webmentions(
        self, resource: str, direction: WebmentionDirection
    ) -> list[Webmention]:
        """
        Retrieve the stored Webmentions for a given URL.

        :param resource: The URL of the resource associated to the Webmentions
        :param direction: The direction of the Webmentions (inbound or outbound)
        :return: A list of Webmentions
        """
        return []

    def mark_sent(self, source: str, target: str) -> None:
        """
        Mark a Webmention as sent.

        :param source: The source URL of the Webmention
        :param target: The target URL of the Webmention
        """
        now = datetime.now(timezone.utc)
        mention = Webmention(
            source=source,
            target=target,
            direction=WebmentionDirection.OUT,
        )
        mention.published = mention.published or now
        mention.created_at = mention.created_at or now
        mention.updated_at = now
        self.store_webmention(mention)



import logging
import re
from concurrent.futures import ThreadPoolExecutor
from urllib.parse import urljoin, urlparse

# noinspection PyPackageRequirements
from bs4 import BeautifulSoup
import requests

from ..storage import WebmentionsStorage
from .._model import ContentTextFormat, Webmention, WebmentionDirection
from ._common import on_mention_callback_wrapper
from ._constants import DEFAULT_HTTP_TIMEOUT, DEFAULT_USER_AGENT

logger = logging.getLogger(__name__)


class OutgoingWebmentionsProcessor:  # pylint: disable=too-few-public-methods
    """
    Process outgoing Webmentions.

    :param storage: Webmentions storage
    :param user_agent: User agent to use
    :param http_timeout: HTTP timeout
    :param on_mention_processed: Callback to call when a Webmention is processed.
        It should accept a :class:`webmentions.Webmention` object.
    :param on_mention_deleted: Callback to call when a Webmention is deleted.
        It should accept a :class:`webmentions.Webmention` object.
    """

    def __init__(
        self,
        storage: WebmentionsStorage,
        *,
        user_agent: str = DEFAULT_USER_AGENT,
        http_timeout: float = DEFAULT_HTTP_TIMEOUT,
        on_mention_processed=None,
        on_mention_deleted=None,
        **_,
    ):
        self._storage = storage
        self._http_timeout = http_timeout
        self._user_agent = user_agent
        self._on_mention_processed = on_mention_callback_wrapper(on_mention_processed)
        self._on_mention_deleted = on_mention_callback_wrapper(on_mention_deleted)

    def process_outgoing_webmentions(
        self,
        source_url: str,
        *,
        text: str | None = None,
        text_format: ContentTextFormat | None = None,
    ) -> None:
        """
        Process an outgoing Webmention.

        :param source_url: The source URL of the Webmention. Ignored if text is
            provided.
        :param text: The text of the Webmention. If not provided, the source URL
            will be fetched.
        :param text_format: The text format of the Webmention. If not provided,
            it will be inferred from the source URL or text.
        """
        if text is None:
            resp = requests.get(
                source_url,
                timeout=self._http_timeout,
                headers={"User-Agent": self._user_agent},
                allow_redirects=True,
            )
            resp.raise_for_status()
            text = resp.text or ""
            text_format = ContentTextFormat.HTML

        text_format = text_format or ContentTextFormat.TEXT
        new_targets = self._extract_targets(text, text_format)

        try:
            old_targets = {
                webmention.target
                for webmention in self._storage.retrieve_webmentions(
                    source_url, direction=WebmentionDirection.OUT
                )
            }
        except Exception:
            old_targets = set()

        removed = old_targets - new_targets
        added_pool = ThreadPoolExecutor(max_workers=10)
        removed_pool = ThreadPoolExecutor(max_workers=10)

        for target_url in sorted(new_targets):
            if source_url == target_url:
                continue  # Ignore self-mentions

            added_pool.submit(self._notify_added, source_url, target_url)

        for target_url in sorted(removed):
            removed_pool.submit(self._notify_removed, source_url, target_url)

        added_pool.shutdown(wait=True)
        removed_pool.shutdown(wait=True)

    def _notify_added(self, source_url: str, target_url: str):
        """
        Notify a target of an outgoing Webmention.
        """
        try:
            self._notify_target(source_url, target_url)
            self._storage.mark_sent(source_url, target_url)
            self._on_mention_processed(
                Webmention(
                    source=source_url,
                    target=target_url,
                    direction=WebmentionDirection.OUT,
                )
            )
        except Exception as e:
            logger.info(
                "Outgoing Webmention failed (source=%s target=%s): %s",
                source_url,
                target_url,
                str(e),
            )

    def _notify_removed(self, source_url: str, target_url: str):
        """
        Notify a target of a removed outgoing Webmention.
        """
        try:
            self._notify_target(source_url, target_url)
            self._storage.delete_webmention(
                source_url, target_url, direction=WebmentionDirection.OUT
            )

            self._on_mention_deleted(
                Webmention(
                    source=source_url,
                    target=target_url,
                    direction=WebmentionDirection.OUT,
                )
            )
        except Exception as e:
            logger.info(
                "Outgoing Webmention deletion failed (source=%s target=%s): %s",
                source_url,
                target_url,
                str(e),
            )

    def _extract_targets(self, text: str, text_format: ContentTextFormat) -> set[str]:
        """
        Given a text and a text format, extract the target URLs from the text.
        """
        if text_format == ContentTextFormat.HTML:
            return self._extract_urls_from_html(text)
        return self._extract_urls_from_markdown_or_text(text)

    def _extract_urls_from_html(self, html: str) -> set[str]:
        """
        Extract URLs from HTML.
        """
        urls: set[str] = set()
        soup = BeautifulSoup(html or "", "html.parser")
        for tag in soup.find_all(["a", "link"]):
            href = tag.get("href")
            if not href:
                continue
            href = str(href).strip()
            if href.lower().startswith("http://") or href.lower().startswith(
                "https://"
            ):
                urls.add(href)
        return self._clean_and_filter_targets(urls)

    def _extract_urls_from_markdown_or_text(self, md: str) -> set[str]:
        """
        Extract URLs from Markdown or text.
        """
        urls: set[str] = set()

        for m in re.finditer(r"]\((https?://[^\s)]+)\)", md, flags=re.IGNORECASE):
            urls.add(m.group(1))

        for m in re.finditer(r"<(https?://[^>\s]+)>", md, flags=re.IGNORECASE):
            urls.add(m.group(1))

        for m in re.finditer(r"(https?://[^\s)\]}>\"']+)", md, flags=re.IGNORECASE):
            urls.add(m.group(1))

        return self._clean_and_filter_targets(urls)

    def _clean_and_filter_targets(self, urls: set[str]) -> set[str]:
        """
        Clean and filter URLs.
        """
        cleaned: set[str] = set()
        for u in urls:
            u2 = u.strip().rstrip('.,;:!?)"]\'"')
            if not u2:
                continue
            cleaned.add(u2)

        return {
            u
            for u in cleaned
            if urlparse(u).scheme in ("http", "https") and urlparse(u).netloc
        }

    def _notify_target(self, source_url: str, target_url: str) -> None:
        """
        Notify a target of an outgoing Webmention.
        """
        endpoint = self._discover_webmention_endpoint(target_url)
        if not endpoint:
            return

        resp = requests.post(
            endpoint,
            data={"source": source_url, "target": target_url},
            timeout=self._http_timeout,
            headers={"User-Agent": self._user_agent},
            allow_redirects=True,
        )

        if resp.status_code >= 400:
            resp.raise_for_status()

    def _discover_webmention_endpoint(self, target_url: str) -> str | None:
        """
        Discover a Webmention endpoint for a target URL.
        """
        resp = requests.get(
            target_url,
            timeout=self._http_timeout,
            headers={"User-Agent": self._user_agent},
            allow_redirects=True,
        )
        resp.raise_for_status()

        # Check if there is a Link header
        link_header = resp.headers.get("Link")
        if link_header:
            for part in link_header.split(","):
                if "rel=" not in part.lower():
                    continue
                if "webmention" not in part.lower():
                    continue
                m = re.search(r"<([^>]+)>", part)
                if m:
                    return urljoin(resp.url, m.group(1))

        # Check if there is a <link> or <a> tag
        soup = BeautifulSoup(resp.text or "", "html.parser")
        for tag in soup.find_all(["link", "a"]):
            rel = tag.get("rel")
            href: str = tag.get("href")  # type: ignore
            if not href:
                continue

            rel_str = " ".join(rel) if isinstance(rel, list) else (rel or "")
            if "webmention" in rel_str.lower():
                return urljoin(resp.url, href)

        return None



from dataclasses import asdict
from datetime import datetime, timezone

import sqlalchemy as sa

from ...._model import (
    Webmention,
    WebmentionDirection,
    WebmentionType,
    WebmentionStatus,
)


class DbWebmention:
    """
    SQLAlchemy base model for Webmentions.

    This model should be inherited by an actual SQLAlchemy model.
    """

    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    source = sa.Column(sa.String, nullable=False)
    target = sa.Column(sa.String, nullable=False)
    direction = sa.Column(
        sa.Enum(WebmentionDirection, name="webmention_direction"), nullable=False
    )
    title = sa.Column(sa.String)
    excerpt = sa.Column(sa.String)
    content = sa.Column(sa.String)
    author_name = sa.Column(sa.String)
    author_url = sa.Column(sa.String)
    author_photo = sa.Column(sa.String)
    published = sa.Column(sa.DateTime, nullable=False)
    status = sa.Column(
        sa.Enum(WebmentionStatus, name="webmention_status"),
        nullable=False,
        default=WebmentionStatus.CONFIRMED,
    )
    mention_type = sa.Column(
        sa.Enum(WebmentionType, name="webmention_type"), nullable=False
    )
    meta = sa.Column(sa.JSON, nullable=False, default={})
    created_at = sa.Column(sa.DateTime, nullable=False)
    updated_at = sa.Column(sa.DateTime, nullable=False)

    # Unique constraint on <source, target, direction>
    __table_args__ = (
        sa.UniqueConstraint(
            "source", "target", "direction", name="uix_source_target_direction"
        ),
    )

    def __init__(self, *_, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    @classmethod
    def columns(cls) -> set[str]:
        return {c.name for c in cls.__table__.columns}  # type: ignore

    @classmethod
    def from_webmention(cls, webmention: Webmention) -> "DbWebmention":
        created_at = webmention.created_at
        if created_at is None:
            created_at = webmention.published
        if created_at is None:
            created_at = datetime.now(timezone.utc)

        columns = cls.columns()
        data = {k: v for k, v in asdict(webmention).items() if k in columns}
        return cls(
            **{
                **data,
                "meta": webmention.metadata or {},
                "created_at": created_at,
                "updated_at": datetime.now(timezone.utc),
            }
        )

    # noinspection PyTypeChecker
    def to_webmention(self) -> Webmention:
        return Webmention(
            source=self.source,  # type: ignore
            target=self.target,  # type: ignore
            direction=WebmentionDirection(self.direction),  # type: ignore
            title=self.title,  # type: ignore
            excerpt=self.excerpt,  # type: ignore
            content=self.content,  # type: ignore
            author_name=self.author_name,  # type: ignore
            author_url=self.author_url,  # type: ignore
            author_photo=self.author_photo,  # type: ignore
            published=self.published,  # type: ignore
            mention_type=(
                WebmentionType.from_raw(self.mention_type)  # type: ignore
                if self.mention_type  # type: ignore
                else WebmentionType.MENTION
            ),
            metadata=dict(self.meta),  # type: ignore
            created_at=self.created_at,  # type: ignore
            updated_at=self.updated_at,  # type: ignore
        )



from dataclasses import asdict, dataclass, field
from datetime import datetime, timezone
from enum import Enum


class WebmentionDirection(str, Enum):
    """
    Enum representing the direction of a Webmention
    (incoming or outgoing).
    """

    IN = "incoming"
    OUT = "outgoing"

    @classmethod
    def from_raw(cls, raw: str) -> "WebmentionDirection":
        try:
            return cls(raw.strip().lower())
        except ValueError:
            value = getattr(cls, raw.strip().upper(), None)
            if not value:
                raise ValueError(f"Unknown direction: {raw}")
            return value


class WebmentionStatus(str, Enum):
    """
    Enum representing the status of a Webmention
    (pending, confirmed, or deleted).
    """

    PENDING = "pending"
    CONFIRMED = "confirmed"
    DELETED = "deleted"


class ContentTextFormat(str, Enum):
    """
    Supported content text formats.
    """

    HTML = "html"
    MARKDOWN = "markdown"
    TEXT = "text"


class WebmentionType(str, Enum):
    """
    Enum representing the type of Webmention.

    Note that this list is not exhaustive, and the
    Webmention recommendation itself does not provide
    any static list.

    This is however a lis of commonly supported types
    in Microformats.
    """

    UNKNOWN = "unknown"
    MENTION = "mention"
    REPLY = "reply"
    LIKE = "like"
    REPOST = "repost"
    BOOKMARK = "bookmark"
    RSVP = "rsvp"
    FOLLOW = "follow"

    @classmethod
    def from_raw(cls, raw: str | None) -> "WebmentionType":
        if not raw:
            return cls.UNKNOWN

        normalized = raw.strip().lower()
        aliases = {
            "in-reply-to": cls.REPLY,
            "reply": cls.REPLY,
            "like-of": cls.LIKE,
            "like": cls.LIKE,
            "repost-of": cls.REPOST,
            "repost": cls.REPOST,
            "bookmark-of": cls.BOOKMARK,
            "bookmark": cls.BOOKMARK,
            "rsvp": cls.RSVP,
            "follow-of": cls.FOLLOW,
            "follow": cls.FOLLOW,
            "mention": cls.MENTION,
        }

        return aliases.get(normalized, cls.UNKNOWN)


@dataclass
class Webmention:
    """
    Data class representing a Webmention.
    """

    source: str
    target: str
    direction: WebmentionDirection
    title: str | None = None
    excerpt: str | None = None
    content: str | None = None
    author_name: str | None = None
    author_url: str | None = None
    author_photo: str | None = None
    published: datetime | None = None
    status: WebmentionStatus = WebmentionStatus.CONFIRMED
    mention_type: WebmentionType = WebmentionType.UNKNOWN
    mention_type_raw: str | None = None
    metadata: dict = field(default_factory=dict)
    created_at: datetime | None = None
    updated_at: datetime | None = None

    def to_dict(self) -> dict:
        """
        :return: A dictionary representation of the Webmention
        """

        def _normalize(value):
            if value is None:
                return None
            if isinstance(value, datetime):
                return value.isoformat()
            if isinstance(value, Enum):
                return value.value
            if isinstance(value, list):
                return [_normalize(v) for v in value]
            if isinstance(value, tuple):
                return [_normalize(v) for v in value]
            if isinstance(value, dict):
                return {k: _normalize(v) for k, v in value.items()}
            return value

        return _normalize(asdict(self))

    def __hash__(self):
        """
        :return: A hash value based on the source, target, and direction.
        """
        return hash((self.source, self.target, self.direction))

    @classmethod
    def build(
        cls, data: dict, direction: WebmentionDirection = WebmentionDirection.IN
    ) -> "Webmention":
        assert data.get("source"), "source is required"
        assert data.get("target"), "target is required"
        mention_type: WebmentionType = (
            data.get("mention_type") or WebmentionType.MENTION
        )

        if isinstance(mention_type, str):
            mention_type = WebmentionType.from_raw(mention_type)

        def _parse_dt(value: object) -> datetime | None:
            dt = None
            if value is None:
                return None
            if isinstance(value, datetime):
                return value

            if isinstance(value, str) and value.strip():
                dt = datetime.fromisoformat(value)
            if isinstance(value, (int, float)):
                dt = datetime.fromtimestamp(value, tz=timezone.utc)

            if dt and dt.tzinfo is None:
                dt = dt.replace(tzinfo=timezone.utc)

            return dt

        published = _parse_dt(data.get("published"))
        created_at = _parse_dt(data.get("created_at"))
        updated_at = _parse_dt(data.get("updated_at"))

        return cls(
            **{
                **{k: v for k, v in data.items() if k in cls.__dataclass_fields__},
                "direction": direction,
                "status": data.get("status") or WebmentionStatus.CONFIRMED,
                "mention_type": mention_type,
                "mention_type_raw": mention_type.value,
                "published": published,
                "created_at": created_at,
                "updated_at": updated_at,
            }
        )



def webmention_link_header_value(endpoint: str) -> str:
    return f'<{endpoint}>; rel="webmention"'


def append_link_header(existing: str | None, to_add: str) -> str:
    if not existing:
        return to_add
    if to_add in existing:
        return existing
    return f"{existing}, {to_add}"



import logging
from typing import Callable

from .._model import Webmention

logger = logging.getLogger(__name__)


def on_mention_callback_wrapper(
    callback: Callable | None,
) -> Callable[[Webmention], None]:
    """
    Wrap a mention callback to catch exceptions and log them.
    """

    def wrapper(mention: Webmention):
        if callback:
            try:
                callback(mention)
            except Exception as e:
                logger.error(
                    "Error on mention processed callback %s: <source=%s target=%s direction=%s>: %s",
                    callback,
                    mention.source,
                    mention.target,
                    mention.direction.value,
                    str(e),
                )

    return wrapper



import logging
import os
import queue
import threading
import time
from typing import Callable

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer
from webmentions import ContentTextFormat

from ._model import ContentChange, ContentChangeType

logger = logging.getLogger(__name__)


class FileSystemWatcher:
    """
    Watches for changes to the content in the given root directory.

    :param root_dir: The root directory to watch.
    :param on_change: The function to call when a change is detected.
        It should accept a :class:`.ContentChange` instance as its only
        argument.
    :param extensions: A tuple of file extensions to watch. Default: all
        text, HTML, and Markdown files.
    :param throttle_seconds: The minimum number of seconds between
        processing changes, to prevent too many consecutive calls when a
        file is written multiple times. Default: 2.
    """

    def __init__(
        self,
        root_dir: str,
        on_change: Callable[[ContentChange], None],
        *,
        extensions: tuple[str, ...] = (".md", ".markdown", ".txt", ".html", ".htm"),
        throttle_seconds: float = 2.0,
    ):
        self.root_dir = root_dir
        self.on_change = on_change
        self.extensions = tuple(e.lower() for e in extensions)
        self._watch_observer = None
        self._watch_queue: "queue.Queue[tuple[str, str]]" = queue.Queue()
        self._watch_stop_event = threading.Event()
        self._watch_thread: threading.Thread | None = None
        self._pending_paths: set[str] = set()
        self._last_event_at: dict[str, float] = {}
        self._throttle_seconds = throttle_seconds
        self._last_processed_at = 0.0
        self._last_event_type: dict[str, str] = {}

    def start(self) -> None:
        """
        Starts the watcher.
        """
        if self._watch_observer is not None:
            return

        if not os.path.isdir(self.root_dir):
            return

        class _EventHandler(FileSystemEventHandler):
            def __init__(self, *_, enqueue_fs_event: Callable[..., None], **__):
                self._enqueue_fs_event = enqueue_fs_event

            def on_created(self, event):
                if getattr(event, "is_directory", False):
                    return
                self._enqueue_fs_event("created", getattr(event, "src_path", ""))

            def on_modified(self, event):
                if getattr(event, "is_directory", False):
                    return
                self._enqueue_fs_event("modified", getattr(event, "src_path", ""))

            def on_deleted(self, event):
                if getattr(event, "is_directory", False):
                    return
                self._enqueue_fs_event("deleted", getattr(event, "src_path", ""))

            def on_moved(self, event):
                if getattr(event, "is_directory", False):
                    return
                src = getattr(event, "src_path", "")
                dest = getattr(event, "dest_path", "")
                self._enqueue_fs_event("deleted", src)
                self._enqueue_fs_event("created", dest)

        self._watch_stop_event.clear()
        self._watch_observer = Observer()
        self._watch_observer.daemon = True
        self._watch_observer.schedule(
            _EventHandler(enqueue_fs_event=self._enqueue_fs_event),
            self.root_dir,
            recursive=True,
        )
        self._watch_observer.start()

        self._watch_thread = threading.Thread(
            target=self._watch_loop,
            name="webmentions-watch-loop",
            daemon=True,
        )
        self._watch_thread.start()
        logger.info("Webmentions file watcher started on %s", self.root_dir)

    def stop(self) -> None:
        """
        Stops the watcher.
        """
        self._watch_stop_event.set()

        if self._watch_observer is not None:
            try:
                self._watch_observer.stop()
                self._watch_observer.join(timeout=5)
            except Exception:
                pass
            self._watch_observer = None

        self._watch_thread = None
        logger.info("Webmentions file watcher stopped")

    def _enqueue_fs_event(self, event_type: str, path: str) -> None:
        if not path:
            return

        abs_path = os.path.abspath(path)
        if not self._is_candidate_path(abs_path):
            return

        self._watch_queue.put((event_type, abs_path))

    def _is_candidate_path(self, abs_path: str) -> bool:
        try:
            root = os.path.abspath(self.root_dir)
            if not (
                abs_path.startswith(root.rstrip(os.sep) + os.sep) or abs_path == root
            ):
                return False
        except Exception:
            return False

        _, ext = os.path.splitext(abs_path)
        return ext.lower() in self.extensions

    def _watch_loop(self) -> None:
        while not self._watch_stop_event.is_set():
            try:
                event_type, path = self._watch_queue.get(timeout=0.5)
            except queue.Empty:
                self._flush_debounced()
                continue

            now = time.monotonic()
            self._pending_paths.add(path)
            self._last_event_at[path] = now
            self._last_event_type[path] = event_type
            self._flush_debounced()

    def _flush_debounced(self) -> None:
        if not self._pending_paths:
            return

        now = time.monotonic()
        if now - self._last_processed_at < self._throttle_seconds:
            return

        ready = [
            p
            for p in list(self._pending_paths)
            if now - self._last_event_at.get(p, now) >= self._throttle_seconds
        ]

        if not ready:
            return

        self._last_processed_at = now
        for p in ready:
            self._pending_paths.discard(p)
            self._last_event_at.pop(p, None)
            event_type = self._last_event_type.pop(p, "")
            change = self._build_change(event_type, p)
            if change is None:
                continue
            try:
                self.on_change(change)
            except Exception as e:
                logger.info("Webmentions change handler failed for %s: %s", p, str(e))

    def _build_change(self, event_type: str, abs_path: str) -> ContentChange | None:
        if event_type == "deleted" or not os.path.isfile(abs_path):
            return ContentChange(
                change_type=ContentChangeType.DELETED,
                path=abs_path,
                text=None,
                text_format=None,
            )

        if event_type == "created":
            change_type = ContentChangeType.ADDED
        else:
            change_type = ContentChangeType.EDITED

        try:
            with open(abs_path, "r", encoding="utf-8") as f:
                text = f.read()
        except Exception:
            text = None

        text_format = self._guess_text_format(abs_path)
        if not text_format:
            return None

        return ContentChange(
            change_type=change_type,
            path=abs_path,
            text=text,
            text_format=text_format,
        )

    @staticmethod
    def _guess_text_format(path: str) -> ContentTextFormat | None:
        _, ext = os.path.splitext(path.lower())
        if ext in (".html", ".htm"):
            return ContentTextFormat.HTML
        if ext in (".md", ".markdown"):
            return ContentTextFormat.MARKDOWN
        if ext in (".txt", ".text"):
            return ContentTextFormat.TEXT

        return None



class WebmentionException(Exception):
    """
    Base exception for webmention errors.
    """

    def __init__(self, source: str | None, target: str | None, message: str, **_):
        self.source = source
        self.target = target
        self.message = message
        super().__init__(f"Invalid Webmention from '{source}' to '{target}': {message}")


class WebmentionGone(WebmentionException, ValueError):
    """
    Exception for webmentions that no longer exist.
    """

    def __init__(
        self, source: str | None, target: str | None, msg: str | None = None, **_
    ):
        super().__init__(source, target, msg or "Webmention no longer exists")



import logging
from datetime import datetime, timezone
from typing import Any

from ..storage import WebmentionsStorage
from .._exceptions import WebmentionException, WebmentionGone
from .._model import Webmention, WebmentionDirection, WebmentionStatus
from ._common import on_mention_callback_wrapper
from ._constants import DEFAULT_HTTP_TIMEOUT, DEFAULT_USER_AGENT
from ._parser import WebmentionsRequestParser

logger = logging.getLogger(__name__)


class IncomingWebmentionsProcessor:  # pylint: disable=too-few-public-methods
    """
    Incoming Webmentions processor.
    """

    def __init__(
        self,
        storage: WebmentionsStorage,
        *,
        base_url: str | None = None,
        http_timeout: float = DEFAULT_HTTP_TIMEOUT,
        user_agent: str = DEFAULT_USER_AGENT,
        on_mention_processed=None,
        on_mention_deleted=None,
        init_mention_status: WebmentionStatus = WebmentionStatus.CONFIRMED,
        **_,
    ):
        self.parser = WebmentionsRequestParser(
            base_url=base_url, http_timeout=http_timeout, user_agent=user_agent
        )
        self._storage = storage
        self._on_mention_processed = on_mention_callback_wrapper(on_mention_processed)
        self._on_mention_deleted = on_mention_callback_wrapper(on_mention_deleted)
        self._initial_mention_status = init_mention_status

    def process_incoming_webmention(
        self, source: str | None, target: str | None
    ) -> Any:
        """
        Process an incoming Webmention.

        :param source: The source URL of the Webmention
        :param target: The target URL of the Webmention
        """
        if source is None or target is None:
            return None

        logger.info("Received Webmention from '%s' to '%s'", source, target)

        try:
            mention = self.parser.parse(source, target)
            assert source and target  # for mypy
        except WebmentionGone:
            assert source and target  # for mypy
            self._storage.delete_webmention(
                source, target, direction=WebmentionDirection.IN
            )

            self._on_mention_deleted(
                Webmention(
                    source=source, target=target, direction=WebmentionDirection.IN
                )
            )

            logger.info("Deleted Webmention from '%s' to '%s'", source, target)
            return None
        except ValueError as e:
            raise WebmentionException(source, target, str(e)) from e

        now = datetime.now(timezone.utc)
        mention.created_at = mention.created_at or mention.published or now
        mention.updated_at = mention.updated_at or now
        mention.status = self._initial_mention_status
        ret = self._storage.store_webmention(mention)
        self._on_mention_processed(mention)
        logger.info("Processed Webmention from '%s' to '%s'", source, target)
        return ret



import logging
import time
from pathlib import Path
from threading import RLock
from typing import Callable

from ....handlers import WebmentionsHandler
from ._model import ContentChange, ContentChangeType
from ._watcher import FileSystemWatcher

logger = logging.getLogger(__name__)


class FileSystemMonitor:
    """
    Encapsulates :class:`.FilesystemWatcher` to watch a directory for
    changes to files and maps file system events to outgoing Webmention
    requests.

    :param handler: The Webmentions handler to use to dispatch
        outgoing Webmentions.
    :param root_dir: The root directory to watch.
    :param file_to_url_mapper: A function that maps file paths to URLs
        to be used in outgoing Webmentions.
    :param extensions: A tuple of file extensions to watch. Default: all
        text, HTML, and Markdown files.
    :param throttle_seconds: The minimum number of seconds between
        processing changes, to prevent too many consecutive calls when a
        file is written multiple times. Default: 2.
    """

    def __init__(
        self,
        handler: WebmentionsHandler,
        root_dir: str,
        file_to_url_mapper: Callable[[str], str],
        *,
        extensions: tuple[str, ...] = (".md", ".markdown", ".txt", ".html", ".htm"),
        throttle_seconds: float = 2.0,
    ):
        self._webmentions_handler = handler
        self._source_url_from_content_path = file_to_url_mapper
        self._watcher: FileSystemWatcher | None = None
        self._root_dir = Path(root_dir).expanduser().resolve()
        self.extensions = tuple(e.lower() for e in extensions)
        self.throttle_seconds = throttle_seconds
        self._watcher_lock = RLock()

    def _on_filesystem_change(self, change: ContentChange):
        if not self._webmentions_handler:
            return

        source_url = self._source_url_from_content_path(change.path)
        t_start = time.monotonic()

        if change.change_type == ContentChangeType.DELETED:
            logger.info(
                "Content deleted, processing outgoing Webmentions: %s",
                source_url,
            )
            self._webmentions_handler.process_outgoing_webmentions(
                source_url,
                text="",
                text_format=change.text_format,
            )
        else:
            logger.info(
                "Content changed, processing outgoing Webmentions: %s",
                source_url,
            )
            self._webmentions_handler.process_outgoing_webmentions(
                source_url,
                text=change.text,
                text_format=change.text_format,
            )

        logger.info(
            "Processed outgoing Webmentions for %s in %.2f seconds",
            source_url,
            time.monotonic() - t_start,
        )

    def start(self) -> None:
        """
        Starts the filesystem watcher.
        """
        with self._watcher_lock:
            if self._watcher is not None:
                return

            self._watcher = FileSystemWatcher(
                root_dir=str(self._root_dir),
                on_change=self._on_filesystem_change,
                extensions=self.extensions,
                throttle_seconds=self.throttle_seconds,
            )

            self._watcher.start()

    def stop(self) -> None:
        """
        Stops the filesystem watcher.
        """
        with self._watcher_lock:
            if self._watcher is None:
                return

            self._watcher.stop()
            self._watcher = None

    def __enter__(self):
        """
        Starts the filesystem watcher.
        """
        self.start()
        return self

    def __exit__(self, *_, **__):
        """
        Stops the filesystem watcher.
        """
        self.stop()



import fastapi as fastapi_upstream  # pylint: disable=W0406

if getattr(fastapi_upstream, "__file__", None) == __file__:
    raise RuntimeError(
        "Local module name 'fastapi.py' is shadowing the upstream 'fastapi' dependency. "
        "Do not run this file directly; import it as 'webmentions.server.adapters.fastapi'."
    )

FastAPI = fastapi_upstream.FastAPI
Form = fastapi_upstream.Form
HTTPException = fastapi_upstream.HTTPException
Query = fastapi_upstream.Query

from ...handlers import WebmentionsHandler
from ..._exceptions import WebmentionException
from ..._model import WebmentionDirection
from ._common import append_link_header, webmention_link_header_value


def bind_webmentions(
    app: FastAPI, handler: "WebmentionsHandler", route: str = "/webmentions"
):
    """
    Bind a FastAPI endpoint to process incoming Webmentions.

    :param app: The FastAPI application to bind the endpoint to.
    :param handler: The WebmentionsHandler to use for processing incoming Webmentions.
    :param route: The route to bind the endpoint to.
    """

    if not hasattr(app.state, "_webmentions_endpoints"):
        app.state._webmentions_endpoints = set()

    app.state._webmentions_endpoints.add(route)

    if not getattr(app.state, "_webmentions_link_header_middleware_installed", False):
        app.state._webmentions_link_header_middleware_installed = True

        @app.middleware("http")
        async def _webmentions_link_header_middleware(request, call_next):
            response = await call_next(request)
            content_type = response.headers.get("content-type")
            if content_type is not None and content_type.split(";", 1)[
                0
            ].strip().startswith("text/"):
                existing = response.headers.get("link")
                for endpoint in getattr(app.state, "_webmentions_endpoints", set()):
                    existing = append_link_header(
                        existing, webmention_link_header_value(endpoint)
                    )
                if existing is not None:
                    response.headers["link"] = existing
            return response

    @app.post(route)
    def webmention(
        source: str | None = Form(default=None),
        target: str | None = Form(default=None),
    ):
        try:
            handler.process_incoming_webmention(source, target)
        except WebmentionException as e:
            raise HTTPException(status_code=400, detail=str(e)) from e

        return {"status": "ok"}

    @app.get(route)
    def retrieve_webmentions(
        resource: str = Query(...),
        direction: WebmentionDirection = Query(...),
    ):
        try:
            direction = WebmentionDirection.from_raw(direction)
            stored = handler.storage.retrieve_webmentions(resource, direction=direction)
        except Exception as e:
            raise HTTPException(status_code=400, detail=str(e)) from e

        return [wm.to_dict() for wm in stored]

    return webmention



from dataclasses import dataclass
from enum import Enum

from webmentions import ContentTextFormat


class ContentChangeType(str, Enum):
    """
    Content change types.
    """

    ADDED = "added"
    EDITED = "edited"
    DELETED = "deleted"


@dataclass(frozen=True)
class ContentChange:
    """
    Content change model.
    """

    change_type: ContentChangeType
    path: str
    text: str | None
    text_format: ContentTextFormat | None



from ._helpers import init_db_storage
from ._model import DbWebmention
from ._storage import DbWebmentionsStorage


__all__ = [
    "DbWebmention",
    "DbWebmentionsStorage",
    "init_db_storage",
]



import logging
import re
from urllib.parse import urlparse
from typing import Any
from datetime import datetime

from bs4 import BeautifulSoup
import mf2py
import requests

from .._exceptions import WebmentionGone
from .._model import Webmention, WebmentionDirection, WebmentionType
from ._constants import DEFAULT_HTTP_TIMEOUT, DEFAULT_USER_AGENT

logger = logging.getLogger(__name__)


class WebmentionsRequestParser:  # pylint: disable=too-few-public-methods
    """
    Parses a Webmention request.
    """

    def __init__(
        self,
        *,
        base_url: str | None = None,
        http_timeout: float = DEFAULT_HTTP_TIMEOUT,
        user_agent: str = DEFAULT_USER_AGENT,
        **_,
    ) -> None:
        self._base_url = base_url
        self._http_timeout = http_timeout
        self._user_agent = user_agent

    def parse(self, source: str | None, target: str | None) -> Webmention:
        """
        Parse a Webmention.

        :param source: The source URL of the webmention
        :param target: The target URL of the webmention
        """
        # Check that both source and target are provided
        if not (source and target):
            raise ValueError(source, target, "Missing source or target URL")

        # Check that the target domain is the same as this server's domain
        if self._base_url:
            target_domain = urlparse(target).netloc
            server_domain = urlparse(self._base_url).netloc
            if target_domain != server_domain:
                raise ValueError("Target URL domain does not match server domain")

        # Check that the source URL is reachable
        resp = requests.get(
            source,
            timeout=self._http_timeout,
            headers={"User-Agent": self._user_agent},
        )

        if resp.status_code in (404, 410):
            raise WebmentionGone(source, target, "Source URL not found")

        resp.raise_for_status()

        # Check that the target URL is included in the source content.
        # The Webmention REC requires an exact match of the target URL.
        if not self._source_mentions_target(resp.text, target):
            raise WebmentionGone(
                source, target, "Target URL not found in source content"
            )

        mention = Webmention(
            source=source,
            target=target,
            direction=WebmentionDirection.IN,
        )

        self._parse_source_payload(mention, resp.text, source, target)
        return mention

    @staticmethod
    def _source_mentions_target(source_body: str, target_url: str) -> bool:
        try:
            soup = BeautifulSoup(source_body, "html.parser")
        except Exception:
            soup = None

        if soup is not None:
            for tag in soup.find_all(href=True):  # type: ignore[arg-type]
                href = tag.get("href")
                if href == target_url:
                    return True
            for tag in soup.find_all(src=True):  # type: ignore[arg-type]
                src = tag.get("src")
                if src == target_url:
                    return True

        return target_url in source_body

    @classmethod
    def _parse_source_payload(
        cls, mention: Webmention, html: str, source_url: str, target_url: str
    ) -> None:
        entry = cls._extract_h_entry(html, source_url)
        if entry:
            cls._fill_from_h_entry(mention, entry, target_url)

        cls._fill_from_html_fallbacks(mention, html)

        if not mention.excerpt and mention.content:
            excerpt = re.sub(r"\s+", " ", mention.content).strip()
            mention.excerpt = excerpt[:240] if excerpt else None

    @staticmethod
    def _extract_h_entry(html: str, source_url: str) -> dict | None:
        try:
            parsed: dict = mf2py.parse(doc=html, url=source_url)  # type: ignore
        except Exception:
            return None

        items = parsed.get("items") or []
        for item in items:
            if "h-entry" in (item.get("type") or []):
                return item

        for item in items:
            for child in item.get("children") or []:
                if "h-entry" in (child.get("type") or []):
                    return child

        return None

    @staticmethod
    def _first_str(  # pylint: disable=too-many-return-statements
        value: Any,
    ) -> str | None:
        if value is None:
            return None
        if isinstance(value, str):
            return value

        if isinstance(value, dict):
            for key in ("value", "url"):
                v = value.get(key)
                if isinstance(v, str) and v:
                    return v
            return None

        if isinstance(value, list) and value:
            v0 = value[0]
            if isinstance(v0, str):
                return v0
            if isinstance(v0, dict):
                for key in ("value", "url"):
                    v = v0.get(key)
                    if isinstance(v, str) and v:
                        return v
            return None

        return None

    @classmethod
    def _extract_author(
        cls, properties: dict
    ) -> tuple[str | None, str | None, str | None]:
        """
        Extract the author of a Webmention.

        :param properties: The properties of the Webmention
        :return: A tuple of (name, url, photo)
        """

        author = properties.get("author")
        if not author:
            return None, None, None

        if isinstance(author, list) and author:
            author = author[0]

        if isinstance(author, str):
            return None, author, None

        if isinstance(author, dict):
            props = author.get("properties") or {}
            name = cls._first_str(props.get("name"))
            url = cls._first_str(props.get("url"))
            photo = cls._first_str(props.get("photo"))
            return name, url, photo

        return None, None, None

    @classmethod
    def _extract_location(cls, properties: dict) -> dict | None:
        location = properties.get("location")
        if not location:
            return None

        if isinstance(location, list) and location:
            location = location[0]

        if isinstance(location, str):
            return {"name": None, "url": location}

        if isinstance(location, dict):
            props = location.get("properties") or {}
            return {
                "type": location.get("type"),
                "name": cls._first_str(props.get("name")),
                "url": cls._first_str(props.get("url")),
                "latitude": cls._first_str(props.get("latitude")),
                "longitude": cls._first_str(props.get("longitude")),
            }

        return None

    @classmethod
    def _fill_from_h_entry(  # pylint: disable=too-many-branches
        cls, mention: Webmention, entry: dict, target_url: str
    ) -> None:
        """
        Fill a Webmention from an h-entry.

        :param mention: The Webmention to fill
        :param entry: The h-entry to fill the Webmention from
        :param target_url: The target URL of the Webmention
        """
        props = entry.get("properties") or {}
        cls._fill_mf2_metadata(mention, entry, props)
        cls._fill_core_fields_from_entry(mention, props)
        cls._fill_author_from_entry(mention, props)
        cls._infer_mention_type_from_entry(mention, props, target_url)
        cls._fill_comments_from_entry(mention, props)

    @classmethod
    def _fill_mf2_metadata(cls, mention: Webmention, entry: dict, props: dict) -> None:
        mention.metadata.setdefault("mf2", {})
        mention.metadata["mf2"]["type"] = entry.get("type")
        mention.metadata["mf2"]["url"] = cls._first_str(props.get("url"))
        mention.metadata["mf2"]["uid"] = cls._first_str(props.get("uid"))
        mention.metadata["mf2"]["category"] = props.get("category") or []
        mention.metadata["mf2"]["syndication"] = props.get("syndication") or []
        mention.metadata["mf2"]["rsvp"] = cls._first_str(props.get("rsvp"))
        mention.metadata["mf2"]["bookmark_of"] = props.get("bookmark-of") or []
        mention.metadata["mf2"]["like_of"] = props.get("like-of") or []
        mention.metadata["mf2"]["repost_of"] = props.get("repost-of") or []
        mention.metadata["mf2"]["in_reply_to"] = props.get("in-reply-to") or []
        mention.metadata["mf2"]["follow_of"] = props.get("follow-of") or []
        mention.metadata["mf2"]["quotation_of"] = props.get("quotation-of") or []
        mention.metadata["mf2"]["photo"] = props.get("photo") or []
        mention.metadata["mf2"]["featured"] = props.get("featured") or []
        mention.metadata["mf2"]["video"] = props.get("video") or []
        mention.metadata["mf2"]["audio"] = props.get("audio") or []
        mention.metadata["mf2"]["location"] = props.get("location") or []
        mention.metadata["mf2"]["photo_url"] = cls._first_str(props.get("photo"))
        mention.metadata["mf2"]["featured_url"] = cls._first_str(props.get("featured"))
        mention.metadata["mf2"]["video_url"] = cls._first_str(props.get("video"))
        mention.metadata["mf2"]["audio_url"] = cls._first_str(props.get("audio"))
        mention.metadata["mf2"]["location_normalized"] = cls._extract_location(props)

    @classmethod
    def _fill_core_fields_from_entry(cls, mention: Webmention, props: dict) -> None:
        mention.title = mention.title or cls._first_str(props.get("name"))
        if not mention.published:
            published = cls._first_str(props.get("published"))
            if published:
                mention.published = datetime.fromisoformat(published)

        summary = cls._first_str(props.get("summary"))
        if summary and not mention.excerpt:
            mention.excerpt = summary

        content = props.get("content")
        if not mention.content and isinstance(content, list) and content:
            c0 = content[0]
            if isinstance(c0, dict):
                mention.content = c0.get("value") or c0.get("html")
            elif isinstance(c0, str):
                mention.content = c0

        if not mention.content:
            mention.content = cls._first_str(props.get("content"))

    @classmethod
    def _fill_author_from_entry(cls, mention: Webmention, props: dict) -> None:
        if mention.author_name or mention.author_url or mention.author_photo:
            return

        name, url, photo = cls._extract_author(props)
        mention.author_name = name
        mention.author_url = url
        mention.author_photo = photo

    @classmethod
    def _infer_mention_type_from_entry(
        cls, mention: Webmention, props: dict, target_url: str
    ) -> None:
        if mention.mention_type != WebmentionType.UNKNOWN:
            return

        like_of = props.get("like-of") or []
        repost_of = props.get("repost-of") or []
        in_reply_to = props.get("in-reply-to") or []
        bookmark_of = props.get("bookmark-of") or []
        follow_of = props.get("follow-of") or []
        rsvp = cls._first_str(props.get("rsvp"))

        if any(target_url == x for x in like_of if isinstance(x, str)):
            mention.mention_type_raw = "like-of"
            mention.mention_type = WebmentionType.from_raw(mention.mention_type_raw)
        elif any(target_url == x for x in repost_of if isinstance(x, str)):
            mention.mention_type_raw = "repost-of"
            mention.mention_type = WebmentionType.from_raw(mention.mention_type_raw)
        elif any(target_url == x for x in bookmark_of if isinstance(x, str)):
            mention.mention_type_raw = "bookmark-of"
            mention.mention_type = WebmentionType.from_raw(mention.mention_type_raw)
        elif any(target_url == x for x in in_reply_to if isinstance(x, str)):
            mention.mention_type_raw = "in-reply-to"
            mention.mention_type = WebmentionType.from_raw(mention.mention_type_raw)
        elif any(target_url == x for x in follow_of if isinstance(x, str)):
            mention.mention_type_raw = "follow-of"
            mention.mention_type = WebmentionType.from_raw(mention.mention_type_raw)
        elif rsvp:
            mention.mention_type_raw = "rsvp"
            mention.mention_type = WebmentionType.from_raw(mention.mention_type_raw)
        else:
            mention.mention_type_raw = "mention"
            mention.mention_type = WebmentionType.from_raw(mention.mention_type_raw)

    @classmethod
    def _fill_comments_from_entry(cls, mention: Webmention, props: dict) -> None:
        comments = props.get("comment") or []
        if comments and "comments" not in mention.metadata:
            mention.metadata["comments"] = cls._extract_comments(comments)

    @classmethod
    def _extract_comments(cls, comments: Any) -> list[dict]:
        if not isinstance(comments, list):
            return []

        extracted: list[dict] = []
        for comment in comments:
            if isinstance(comment, str):
                extracted.append({"url": comment})
                continue

            if not isinstance(comment, dict):
                continue

            c_props = comment.get("properties") or {}
            author_name, author_url, author_photo = cls._extract_author(c_props)
            published = cls._first_str(c_props.get("published"))
            content = None
            c_content = c_props.get("content")
            if isinstance(c_content, list) and c_content:
                c0 = c_content[0]
                if isinstance(c0, dict):
                    content = c0.get("value") or c0.get("html")
                elif isinstance(c0, str):
                    content = c0
            if not content:
                content = cls._first_str(c_props.get("content"))

            extracted.append(
                {
                    "type": comment.get("type"),
                    "name": cls._first_str(c_props.get("name")),
                    "url": cls._first_str(c_props.get("url")),
                    "published": published,
                    "content": content,
                    "author": {
                        "name": author_name,
                        "url": author_url,
                        "photo": author_photo,
                    },
                }
            )

        return extracted

    @staticmethod
    def _fill_from_html_fallbacks(mention: Webmention, html: str) -> None:
        """
        Fill a Webmention from HTML.

        :param mention: The Webmention to fill
        :param html: The HTML to fill the Webmention from
        """
        soup = BeautifulSoup(html, "html.parser")

        if not mention.title:
            og_title: dict = soup.find("meta", attrs={"property": "og:title"})  # type: ignore
            tw_title: dict = soup.find("meta", attrs={"name": "twitter:title"})  # type: ignore
            if og_title and og_title.get("content"):
                mention.title = og_title.get("content")
            elif tw_title and tw_title.get("content"):
                mention.title = tw_title.get("content")
            elif soup.title and soup.title.string:
                mention.title = soup.title.string.strip()

        if not mention.author_name:
            meta_author: dict = soup.find("meta", attrs={"name": "author"})  # type: ignore
            if meta_author and meta_author.get("content"):
                mention.author_name = meta_author.get("content")

        if not mention.published:
            meta_pub: dict = soup.find("meta", attrs={"property": "article:published_time"})  # type: ignore
            if meta_pub and meta_pub.get("content"):
                mention.published = datetime.fromisoformat(meta_pub["content"])

        if not mention.content:
            desc: dict = soup.find("meta", attrs={"property": "og:description"})  # type: ignore
            if desc and desc.get("content"):
                mention.content = desc.get("content")

        if mention.content and not mention.excerpt:
            excerpt = re.sub(r"\s+", " ", mention.content).strip()
            mention.excerpt = excerpt[:250] if excerpt else None
